@@ -6,11 +6,11 @@
 from typing import Optional
 
 # API settings
-_api_key = None          # type: Optional[str]
+_api_key = None  # type: Optional[str]
 _application_key = None  # type: Optional[str]
-_api_version = 'v1'
-_api_host = None         # type: Optional[str]
-_host_name = None        # type: Optional[str]
+_api_version = "v1"
+_api_host = None  # type: Optional[str]
+_host_name = None  # type: Optional[str]
 _hostname_from_config = True
 _cacert = True
 
@@ -49,3 +49,4 @@ from datadog.api.roles import Roles
 from datadog.api.permissions import Permissions
 from datadog.api.service_level_objectives import ServiceLevelObjective
 from datadog.api.synthetics import Synthetics
+from datadog.api.logs import Logs
