@@ -1,13 +1,15 @@
 # Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
 # This product includes software developed at Datadog (https://www.datadoghq.com/).
 # Copyright 2015-Present Datadog, Inc
-try:
-    from collections.abc import Iterable
-except ImportError:
-    from collections import Iterable
 from numbers import Number
+import sys
 import time
 
+if sys.version_info[0] >= 3:
+    from collections.abc import Iterable
+else:
+    from collections import Iterable
+
 
 def format_points(points):
     """
