@@ -43,9 +43,17 @@ def report_warnings(res):
 
 
 class DogshellConfig(IterableUserDict):
-    def load(self, config_file, api_key, app_key):
+    def load(self, config_file, api_key, app_key, api_host):
         config = configparser.ConfigParser()
 
+        if api_host is not None:
+            if api_host in ("us" or "datadoghq.com"):
+                self["api_host"] = "https://datadoghq.com"
+            elif api_host in ("datadoghq.eu", "eu"):
+                self["api_host"] = "https://datadoghq.eu"
+            elif api_host in ("us3.datadoghq.com", "us3"):
+                self["api_host"] = "https://us3.datadoghq.com"
+
         if api_key is not None and app_key is not None:
             self["api_key"] = api_key
             self["app_key"] = app_key
