@@ -7,6 +7,8 @@ import datetime
 import json
 import re
 
+from datadog.util.compat import conditional_lru_cache
+
 TAG_INVALID_CHARS_RE = re.compile(r"[^\w\d_\-:/\.]", re.UNICODE)
 TAG_INVALID_CHARS_SUBS = "_"
 
@@ -29,5 +31,12 @@ def force_to_epoch_seconds(epoch_sec_or_dt):
     return epoch_sec_or_dt
 
 
-def normalize_tags(tag_list):
+@conditional_lru_cache
+def _normalize_tags_with_cache(tag_list):
     return [TAG_INVALID_CHARS_RE.sub(TAG_INVALID_CHARS_SUBS, tag) for tag in tag_list]
+
+
+def normalize_tags(tag_list):
+    # We have to turn our input tag list into a non-mutable tuple for it to
+    # be hashable (and thus usable) by the @lru_cache decorator.
+    return _normalize_tags_with_cache(tuple(tag_list))
